# FILE: mmSolve2D.tcl
#
# This file must be interpreted by the mmsolve shell.

# Support libraries
package require Net
if {[Oc_Main HasTk]} {
    wm withdraw .        ;# "." is just an empty window.
    package require Ow
    Ow_SetIcon .
} else {
    # Set up to write log messages to oommf/app/mmsolve2d/mmsolve2d.errors.
    # First open and truncate this file, and write a process header.
    # Subsequent opens inside the FileReport proc open for append.
    set filename [file join [file dirname $argv0] mmsolve2d.errors]
    if {[catch {open $filename w} chanid]} {
	if {0} {
	    // Disable error message for beta release.  This
	    // should be tied to a global "debug" level.
	    Oc_Log Log "Error in FileReport: unable to\
		    open error log file $filename: $chanid" error
	}
    } else {
	puts $chanid "\[[clock format [clock seconds]]\]\
		    Error log start for process [pid]"
	close $chanid
	# Log handler.  If we drop a lot of messages here,
	# then we should change this to open the file once,
	# and just pass the chanid around.
	proc FileReport { filename msg type src } {
	    catch {
		puts stderr "[pid] $src $type: $msg"
		fflush stderr
	    }
	    if {[catch {open $filename a} chanid]} {
		puts stderr "Error in FileReport: unable to\
			open error log file $filename: $chanid"
	    } else {
		catch { puts $chanid "\[[clock format [clock seconds]]\]\
			[pid] $src $type: $msg" }
		close $chanid
	    }
	}
	Oc_Log SetLogHandler [list FileReport $filename] panic Oc_Log
	Oc_Log SetLogHandler [list FileReport $filename] error Oc_Log
	Oc_Log SetLogHandler [list FileReport $filename] warning Oc_Log
	Oc_Log SetLogHandler [list FileReport $filename] info Oc_Log
	Oc_Log SetLogHandler [list FileReport $filename] panic
	Oc_Log SetLogHandler [list FileReport $filename] error
	Oc_Log SetLogHandler [list FileReport $filename] warning
	Oc_Log SetLogHandler [list FileReport $filename] info
    }
}

set commands {
    Solver	{{LoadProblem RemoteRelayInput ProblemDescription}}
}

array set inputs {
    ProblemDescription	{ProbEd
			 GetProbDescFile
			 LoadProbDescFile
                        }
}
array set events {}
array set outputs {}

proc LoadProbDescFile { mif_file } {
    global commands events inputs outputs
#    catch { unset inputs } ;###
    catch { unset outputs }  ;# "outputs" is _not_ always set!
    catch { unset events }
    catch { unset commands } 
#    array set inputs {}   ;###
# The commands which can be invoked remotely
set commands {
    Solver	{
		 {Reset ResetThread} 
                 {LoadProblem RemoteRelayInput ProblemDescription}
                 {Run ChangeSolverState Run}  
                 {Relax ChangeSolverState Relax}  
                 {Pause ChangeSolverState Pause}  
		 {Field- FieldAdjust -1}
		 {Field+ FieldAdjust 1}
		}
}

# The events generated by this thread.
array set events {
    Iteration	{1 solver GetODEStepCount}
    Equilibrium	{1 solver GetFieldStepCount}
}

# The quantities which this thread produces as output
#	The last entry should be moved to be set by controller
array set outputs {
    Magnetization	{vectorField
			 {Iteration Equilibrium}
			 solver
			 WriteMagFile
			 datafile
                        }
    TotalField 		{vectorField
			 {Iteration Equilibrium}
			 solver
			 WriteFieldFile
			 datafile
                        }
    DataTable		{DataTable
			 {Iteration Equilibrium}
			 ""
			 MakeDataTable
			 DataTable
			}
}


    global solver protocol
    mms_mif New mif
    $mif Read $mif_file
    if {![info exists solver]} {
	# First pass---new solver
	mms_solver New solver $mif
	# Overkill until something better is designed
	foreach c [Net_Connection Instances] {
	    $c Tell InterfaceChange
	}
	em_eventHandler new $solver delete { after 2000 exit }
    } else {
	# Reset existing solver
	ResetThread $mif
    }
    $mif Delete  ;# $solver makes a private copy of $mif
    file delete $mif_file
}

Net_Protocol New protocol -name "OOMMF solver protocol 0.0"
$protocol AddMessage start HasGui {} {
    return [list start [list 0 1]]
}
$protocol AddMessage start Commands {} {
    global commands
    return [list start [list 0 $commands]]
}
$protocol AddMessage start Outputs {} {
    global outputs
    return [list start [list 0 [array get outputs]]]
}
$protocol AddMessage start Inputs {} {
    global inputs
    return [list start [list 0 [array get inputs]]]
}
$protocol AddMessage start Events {} {
    global events
    return [list start [list 0 [array get events]]]
}
$protocol AddMessage start SetHandler {args} {
    return [list start [list 0 [eval SetHandler $args]]]
}
$protocol AddMessage start SetInput {args} {
    return [list start [list 0 [eval SetInput $args]]]
}
$protocol AddMessage start UnsetHandler {args} {
    return [list start [list 0 [eval UnsetHandler $args]]]
}
$protocol AddMessage start RemoteRelayOutput {args} {
    return [list start [list 0 [eval RemoteRelayOutput $args]]]
}
$protocol AddMessage start RemoteRelayInput {args} {
    return [list start [list 0 [eval RemoteRelayInput $args]]]
}
$protocol AddMessage start FieldAdjust {args} {
    return [list start [list 0 [eval FieldAdjust $args]]]
}
$protocol AddMessage start ResetThread {args} {
    return [list start [list 0 [eval ResetThread $args]]]
}
$protocol AddMessage start ChangeSolverState {args} {
    return [list start [list 0 [eval ChangeSolverState $args]]]
}
$protocol AddMessage start delete {} {
    global protocol
    return [list start [list 0 [after 100 $protocol Delete; after 2000 exit]]]
}
Net_Server New server -protocol $protocol -alias mmSolve2D \
        -log [em_extension ApplicationLog]
$server Start 0

# Delay "nice" of this process until any children are spawned.
Net_Account New a
if {[$a Ready]} {
    Oc_MakeNice
} else {   
    Oc_EventHandler New _ $a Ready Oc_MakeNice -oneshot 1
}

proc SetInput {data host acct pid} {
    global sourcethread
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    set sourcethread($data) $t
}
proc SetHandler {event data host acct pid cnt} {
    global events
    UnsetHandler $event $data $host $acct $pid
    global handler outputs
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    set evinfo $events($event)
    set srcName [lindex $evinfo 1]
    upvar #0 $srcName src
    set handler($event,$host,$acct,$pid,$data) [em_eventHandler new $src \
            $event [list GenericHandler $event $data $cnt $t]]
    em_eventHandler new $t delete \
	    [list UnsetHandler $event $data $host $acct $pid] \
	    -oneshot 1
}

proc UnsetHandler {event data host acct pid} {
    global handler
    if {[info exists handler($event,$host,$acct,$pid,$data)]} {
        $handler($event,$host,$acct,$pid,$data) delete
        unset handler($event,$host,$acct,$pid,$data)
    }
}

proc GenericHandler {event data cnt thread} {
    global events
    set evinfo $events($event)
    if {[lindex $evinfo 0]} {
        set srcName [lindex $evinfo 1]
        upvar #0 $srcName src
        set evcount [eval $src [lindex $evinfo 2]]
        if {$evcount % $cnt} {
            return
        }
    }
    RelayOutput $data $thread
}

proc RemoteRelayOutput { data host acct pid } {
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    RelayOutput $data $t
}

# Send the data indicated by 'data' to the process represented by 'thread'
proc RelayOutput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            # Two future events may occur:  'thread' may report that
            # it is ready, or 'thread' may report that it is being
            # deleted.  Set up to handle these events, so that if/when
            # 'thread' becomes ready, we try this routine again.

            # When 'thread' reports it's being deleted, delete all handlers
            # of events from 'thread' created by this proc.  No harm done
            # if this is called when there are no such handlers left, so
            # multiple copies of this handler may exist.
            set h [em_eventHandler new $thread delete \
                    [list em_eventHandler DeleteGroup ${thread}_RelayOutput] \
                    -groups ${thread}_RelayOutput]

            # When 'thread' reports that it is ready, try the Relay Output
            # again.  Then delete this handler (-oneshot) and the corresponding
            # handler for the delete event.
            em_eventHandler new $thread Ready \
                    "RelayOutput $data $thread ; $h delete" \
                    -oneshot 1 -groups ${thread}_RelayOutput
	}
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send data.
        global outputs
        set opinfo $outputs($data)
        set srcName [lindex $opinfo 2]
        if {[string length $srcName]} {
            upvar #0 $srcName src
            set datavalue [eval $src [lindex $opinfo 3]]
        } else {
            set datavalue [eval [lindex $opinfo 3]]
        }
        set method [lindex $opinfo 4]
        $thread Send $method $datavalue
    }
}

proc RemoteRelayInput { data } {
    global sourcethread
    RelayInput $data $sourcethread($data)
}

# Send request for 'data' to the process represented by 'thread'
proc RelayInput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            # Two future events may occur:  'thread' may report that
            # it is ready, or 'thread' may report that it is being
            # deleted.  Set up to handle these events, so that if/when
            # 'thread' becomes ready, we try this routine again.

            # When 'thread' reports it's being deleted, delete all handlers
            # of events from 'thread' created by this proc.  No harm done
            # if this is called when there are no such handlers left, so
            # multiple copies of this handler may exist.
            set h [em_eventHandler new $thread delete \
                    [list em_eventHandler DeleteGroup ${thread}_RelayInput] \
                    -groups ${thread}_RelayInput]

            # When 'thread' reports that it is ready, try the Relay Output
            # again.  Then delete this handler (-oneshot) and the corresponding
            # handler for the delete event.
            em_eventHandler new $thread Ready \
                    "RelayInput $data $thread ; $h delete" \
                    -oneshot 1 -groups ${thread}_RelayInput
	}
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send request.
        global inputs handler
        set ininfo $inputs($data)
        set qid [$thread Send [lindex $ininfo 1]]    
        set handler(input,[$thread Cget -hostname],[$thread Cget -accountname],[$thread Cget -pid],$data) \
               [em_eventHandler new $thread Reply$qid \
               [list ReceiveRelayInputReply $thread $data]]
        # Consider handler for timeout event?
    }
}

proc ReceiveRelayInputReply {thread data} {
    global handler inputs
    set host [$thread Cget -hostname]
    set acct [$thread Cget -accountname]
    set pid [$thread Cget -pid]
    $handler(input,$host,$acct,$pid,$data) delete
    unset handler(input,$host,$acct,$pid,$data)
    set reply [$thread Get]
    if {![lindex $reply 0]} {
        set ininfo $inputs($data)
        eval [lindex $ininfo 2] [lrange $reply 1 end]
    }
}

# Initialize solver
#$solver Configure -relax_stop_mxh 1e-3
set Bfield_multiplier 1000  ;# Convert T to mT

proc MakeDataTable {} {
    global solver Bfield_multiplier
    set table [list [list Iteration "" [$solver GetODEStepCount]]]
    lappend table [list "Field Updates" {} [$solver GetFieldUpdateCount]]
    lappend table [list "Sim Time" "ns" [expr [$solver GetSimulationTime]*1e9]]
    lappend table [list "Time Step" "ns" [expr [$solver GetTimeStep]*1e9]]
    lappend table [list "Step Size" "" [$solver GetStepSize]]
    set Bfield [$solver GetNomAppliedField]
    lappend table [list Bx mT [expr $Bfield_multiplier * [lindex $Bfield 0]]]
    lappend table [list By mT [expr $Bfield_multiplier * [lindex $Bfield 1]]]
    lappend table [list Bz mT [expr $Bfield_multiplier * [lindex $Bfield 2]]]
    set Bmag [expr sqrt([lindex $Bfield 0]*[lindex $Bfield 0] \
	    + [lindex $Bfield 1]*[lindex $Bfield 1] \
	    + [lindex $Bfield 2]*[lindex $Bfield 2])]
    lappend table [list B mT [expr $Bfield_multiplier * $Bmag]]
    lappend table [list "|m x h|" "" [$solver GetMaxMxH]]
    set mag [$solver GetAveMag]
    lappend table [list Mx/Ms "" [lindex $mag 0]]
    lappend table [list My/Ms "" [lindex $mag 1]]
    lappend table [list Mz/Ms "" [lindex $mag 2]]
    set energies [join [$solver GetEnergyDensities]]
    lappend table [list "Total Energy"      "J/m^3" [lindex $energies 1]]
    lappend table [list "Exchange Energy"   "J/m^3" [lindex $energies 3]]
    lappend table [list "Anisotropy Energy" "J/m^3" [lindex $energies 5]]
    lappend table [list "Demag Energy"      "J/m^3" [lindex $energies 7]]
    lappend table [list "Zeeman Energy"     "J/m^3" [lindex $energies 9]]
    lappend table [list "Max Angle" "deg" [$solver GetMaxAngle]]
    return $table
}

proc FieldAdjust { step } {
    global solver 
    if {$step != 0} {
	$solver IncrementFieldStep $step
    }
}

proc ResetThread { {mif {}} } {
    global solver
    $solver Reset $mif
}

proc ChangeSolverState { newstate } {
    global solver
    $solver ChangeSolverState $newstate
}

vwait forever
